program pslst;

{$IOCHECKS OFF}

uses
  Libc, SysUtils, Classes;

type
  TWhichProcesses = (wpNone, wpAll, wpUser, wpNonRoot, wpRunning);

var
  //Are we being invoked to kill a process or list processes
  KillProcess: Boolean;
  //Which processes should we list
  WhichProcesses: TWhichProcesses;
  //Some convenient user name strings
  CurrentUser, Root: String;
  //Directory of process being worked on (/proc/XXX)
  PIDDir: String;
  //PID of process being worked on
  PID: __pid_t;
  //An easy to read version of /proc/XXX/status
  ProcessStatus: TStringList;
  //File search record
  SearchRec: TSearchRec;

{$if Declared(RTLVersion) and (RTLVersion < 14.20)}
const
  //Kylix 1 omission
  _POSIX_PATH_MAX = 1024;
{$ifend}

function GetProcessStatusLine(const Prefix: String): String;
var
  I: Integer;
  Line: String;
begin
  Result := '';
  for I := 0 to ProcessStatus.Count - 1 do
  begin
    Line := ProcessStatus[I];
    if Pos(Prefix, Line) > 0 then
    begin
      Result := Copy(Line, Succ(Length(Prefix)), Length(Line));
      Break;
    end
  end
end;

function GetProcessUser: String;
var
  UID: string;
  TabPos: Integer;
begin
  UID := GetProcessStatusLine('Uid:'#9);
  TabPos := Pos(#9, UID);
  //Typically there are 3 UID values separated by tabs, but check, JIC
  if TabPos > 0 then
    Delete(UID, TabPos, Length(UID))
  else
    UID := Trim(UID);
  Result := getpwuid(StrToInt(UID))^.pw_name
end;

function GetProcessStatus: String;
begin
  Result := Copy(GetProcessStatusLine('State:'#9), 1, 1)
end;

function GetProcessMemVal(const Prefix: String): String;
begin
  Result := GetProcessStatusLine(Prefix);
  //Remove the trailing kB
  SetLength(Result, Length(Result) - 2);
  //Trim off excess white space
  Result := Trim(Result)
end;

function GetProcessCmdLine(Maxlen: Integer): String;
var
  F: file of Char;
  Ch: Char;
begin
  Result := '';
  //The file /proc/XXX/cmdline is a series of zero-terminated strings,
  //possibly followed by an extra zero terminator
  //Note that file size is reported as 0 for kernel memory-mapped files
  AssignFile(F, PIDDir + 'cmdline');
  Reset(F);
  try
    repeat
      Read(F, Ch);
      if Ch <> #0 then //read till a 0 char
        Result := Result + Ch
      else
      begin
        //We will find an extra #0 at EOF in some cases here
        //otherwise we'll trigger a read past EOF IOError
        Read(F, Ch);
        if (IOResult = 0) and (Ch <> #0) then
          Result := Result + #32 + Ch;
      end
    until (Ch = #0) or (IOResult <> 0)
  finally
    CloseFile(F)
  end;
  if Length(Result) > Maxlen then
    Result := Copy(Result, 1, MaxLen - 3) + '...'
end;

procedure LoadTextFile(const Filename: String; List: TStrings);
var
  F: TextFile;
  S: String;
  Stop: Boolean;
begin
  AssignFile(F, FileName);
  Reset(F);
  try
    List.Clear;
    repeat
      ReadLn(F, S);
      Stop := (IOResult <> 0) or (S = '');
      if not Stop then
        List.Add(S)
    until Stop
  finally
    CloseFile(F)
  end;
end;

procedure WriteProcessInfo;
var
  State, User: String;
begin
  LoadTextFile(PIDDir + 'status', ProcessStatus);
  //If user only wants running processes, leave for any other state
  State := GetProcessStatus;
  if (WhichProcesses = wpRunning) and ((Length(State) = 0) or (State[1] <> 'R')) then
    Exit;
  //If user only wants non-root processes, leave if this is a root processes
  User := GetProcessUser;
  if (WhichProcesses = wpNonRoot) and (User = Root) then
    Exit;
  //If user only wants user processes, leave for other users' processes
  if (WhichProcesses = wpUser) and (User <> CurrentUser) then
    Exit;
  WriteLn(Format('%.4d', [PID]), User:10, State:6,
    GetProcessMemVal('VmSize:'#9):6,
    GetProcessMemVal('VmRSS:'#9):6, #32,
    GetProcessCmdLine(47));
end;

procedure ProcessCommandLine;
var
  ListOption: Char;
begin
  //Check for kill request
  if (ParamCount = 2) and (ParamStr(1) = 'k') and (StrToIntDef(ParamStr(2), 0) <> 0) then
  begin
    KillProcess := True;
    PID := StrToInt(ParamStr(2))
  end
  else
  begin
    WhichProcesses := wpNone;
    //Check for process list options (all single-character)
    if (ParamCount = 1) and (Length(ParamStr(1)) = 1) then
    begin
      ListOption := ParamStr(1)[1]; //1st & only char in parameter
      case ListOption of
        'a': WhichProcesses := wpAll;
        'n': WhichProcesses := wpNonRoot;
        'r': WhichProcesses := wpRunning;
        'u': WhichProcesses := wpUser;
      end
    end
  end;
  //Any bad options result in the program help screen
  if not KillProcess and (WhichProcesses = wpNone) then
  begin
    WriteLn('pslst Version 1.0');
    WriteLn('Written in Borland Kylix by Brian Long');
    WriteLn;
    WriteLn('Syntax:');
    WriteLn('  pslst [options]');
    WriteLn;
    WriteLn('options can be one of:');
    WriteLn('  k PID  - kill specified process (PID is decimal Process IDentifier)');
    WriteLn('  a      - list all processes');
    WriteLn('  n      - list processes not owned by root');
    WriteLn('  r      - list only running processes');
    WriteLn('  u      - list only user processes');
  end
end;

begin
  //Default to safe file access mode
  FileMode := fmOpenRead or fmShareDenyNone;
  //Set up root & current user name strings
  Root := getpwuid(0)^.pw_name;
  CurrentUser := getpwuid(getuid)^.pw_name;
  //Set up string list for status files to be read into
  ProcessStatus := TStringList.Create;
  //Check command line parameters and see if we are killing or listing
  ProcessCommandLine;
  if KillProcess then
    kill(PID, SIGKILL)
  else
    if WhichProcesses <> wpNone then
    begin
      //Write headers
      WriteLn(' PID      User State   VSZ   RSS Command');
      //Loop across processes and write out details
      if FindFirst('/proc/*', faAnyFile, SearchRec) = 0 then
        try
          repeat
            if SearchRec.Attr and faDirectory > 0 then
            begin
              PID := StrToIntDef(SearchRec.Name, 0);
              if PID <> 0 then
              begin
                PIDDir := Format('/proc/%s/', [SearchRec.Name]);
                WriteProcessInfo;
              end
            end
          until FindNext(SearchRec) <> 0
        finally
          FindClose(SearchRec)
        end
    end;
  ProcessStatus.Free
end.
